//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "common.h"

//-----------------------------------------------------------------------------
// Helper functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
void prepareNameAttr(WORK_PARAMS *pWP, int iDirEntryNum)
{
  int ii, jj;
  FCB *pFCB;

  pFCB = (FCB *) &pWP->dir[iDirEntryNum];

  pWP->fna.cUser = pFCB->cUser;

  for (ii = 0; ii < 8; ii++) {
    if (pFCB->caFN[ii] == ' ')
      break;
    pWP->fna.szFN[ii] = (pFCB->caFN[ii] & 127);
  }
  if (ii == 0) {
    pWP->fna.szFN[0] = ' ';
    ii++;
  }

  if ((pFCB->caFT[0] & 127) != ' ') {
    pWP->fna.szFN[ii++] = '.';
    for (jj = 0; jj < 3; jj++) {
      if ((pFCB->caFT[jj] & 127) == ' ')
        break;
      pWP->fna.szFN[ii++] = (pFCB->caFT[jj] & 127);
    }
  }

  pWP->fna.szFN[ii] = '\0';
  strupr(pWP->fna.szFN);

  if (pFCB->caFT[0] & 0x80)
    pWP->fna.bReadOnly = TRUE;
  else
    pWP->fna.bReadOnly = FALSE;

  if (pFCB->caFT[1] & 0x80)
    pWP->fna.bHidden = TRUE;
  else
    pWP->fna.bHidden = FALSE;
}
//-----------------------------------------------------------------------------
void prepareWildcard(char *pszDest, const char *pszSrc, int iMax)
{
  int iCnt = 0;

  while (iCnt < iMax) {
    if (pszSrc == NULL || *pszSrc == '\0')
      pszDest[iCnt] = ' ';
    else if (*pszSrc == '*')
      pszDest[iCnt] = '?';
    else
      pszDest[iCnt] = *pszSrc++;
    iCnt++;
  }
  pszDest[iCnt] = '\0';
}
//-----------------------------------------------------------------------------
/*
  - Name can have up to 8 characters.
  - Ext can have up to 3 characters.
  - Name cannot contain space inside.
  - Only printable character in Name and Ext.
*/
BOOL checkFileName(WORK_PARAMS *pWP, char *pszMask, const char *pszFN)
{
  char *p, c;
  char szFNx[13], szName[9], szExt[4];
  int iLen, ii;
  BOOL bSpace;

  iLen = strlen(pszFN);
  if (iLen == 0 || iLen > 12)
    return FALSE;
  strcpy(szFNx, pszFN);
  strupr(szFNx);

  p = strrchr(szFNx, '.');
  if (p != NULL) {
    *p = 0;
    p++;
    if (strlen(p) > 3)
      return FALSE;
  }
  prepareWildcard(szExt, p, 3);

  iLen = strlen(szFNx);
  if (iLen == 0 || iLen > 8)
    return FALSE;
  prepareWildcard(szName, szFNx, 8);

  memcpy(szFNx, szName, 8);
  szFNx[8] = '.';
  memcpy(szFNx + 9, szExt, 3);
  szFNx[12] = 0;

  bSpace = FALSE;
  for (ii = 0; ii < 12; ii++) {
    if (ii == 8)
      continue;

    c = szFNx[ii];
    if (c < ' ' || c > '~')
      return FALSE;
    if (c == ' ') {
      if (pszMask != NULL && ii == 0)
        return FALSE;
      bSpace = TRUE;
    }
    else if (bSpace && ii < 8)
      return FALSE;

    if (pszMask != NULL) {
      if (c == '<' || c == '>' || c == '.' || c == ',' || c == ';' || c == '"'
          || c == ':' || c == '=' || c == '[' || c == ']' || c == '%'
          || c == '|' || c == '(' || c == ')' || c == '/' || c == '\\')
        return FALSE;
    }

    if (pszMask == NULL && (c == '?' || c == '*'))
      return FALSE;
  }

  if (pszMask != NULL) {
    memcpy(pszMask, szName, 8);
    memcpy(pszMask + 8, szExt, 3);

    pWP->bHasWildcard = (strchr(pszMask, '?') == NULL) ? FALSE : TRUE;
  }

  return TRUE;
}
//-----------------------------------------------------------------------------
BOOL markAU(WORK_PARAMS *pWP, int iAUnum, BOOL bState)
{
  if (iAUnum >= pWP->iAllocationUnits)
    return FALSE;

  if (bState)
    pWP->au[iAUnum / 8] |= (BYTE) (1 << (7 - (iAUnum & 7)));
  else
    pWP->au[iAUnum / 8] &= (BYTE) ~(1 << (7 - (iAUnum & 7)));

  return TRUE;
}
//-----------------------------------------------------------------------------
int findFirstFreeAU(WORK_PARAMS *pWP)
{
  int ii, jj, iRol;

  for (ii = 0; ii < pWP->iAllocVectorSize; ii++) {
    iRol = 0x80;
    for (jj = 0; jj < 8; jj++) {
      if ((pWP->au[ii] & iRol) == 0)
        return (ii * 8 + jj);
      iRol >>= 1;
    }
  }

  return pWP->iAllocationUnits;
}
//-----------------------------------------------------------------------------
int getFreeAuCount(WORK_PARAMS *pWP)
{
  int ii, jj, iRol;
  int iFreeAU, iAUnum;

  iFreeAU = 0;
  iAUnum = 0;
  for (ii = 0; ii < pWP->iAllocVectorSize && iAUnum < pWP->iAllocationUnits; ii++) {
    iRol = 0x80;
    for (jj = 0; jj < 8 && iAUnum < pWP->iAllocationUnits; jj++) {
      if ((pWP->au[ii] & iRol) == 0)
        iFreeAU++;
      iRol >>= 1;
      iAUnum++;
    }
  }

  return iFreeAU;
}
//-----------------------------------------------------------------------------
int getFreeDirEntriesCount(WORK_PARAMS *pWP)
{
  int iCnt = 0, ii;

  for (ii = 0; ii < pWP->iDirEntries; ii++) {
    if (pWP->dir[ii].cUser == FILL_BYTE)
      iCnt++;
  }

  return iCnt;
}
//-----------------------------------------------------------------------------
int findFreeDirEntry(WORK_PARAMS *pWP)
{
  int ii;

  for (ii = 0; ii < pWP->iDirEntries; ii++) {
    if (pWP->dir[ii].cUser == FILL_BYTE)
      return ii;
  }

  return -1;
}
//-----------------------------------------------------------------------------
int findDirEntry(WORK_PARAMS *pWP, char *pcaEnt, int iFrom)
{
  int ii;

  for (ii = iFrom; ii < pWP->iDirEntries; ii++) {
    if (memcmp(pcaEnt, &pWP->dir[ii].cUser, 12) == 0)
      return ii;
  }

  return -1;
}
//-----------------------------------------------------------------------------
int calcFileRecs(WORK_PARAMS *pWP, char *pcaEnt)
{
  int iRecs = 0;
  int iFrom = 0;
  int iExt;

  // go through all directory entries of the file
  // last directory entry of the file gives a final value
  while (TRUE) {
    iFrom = findDirEntry(pWP, pcaEnt, iFrom);
    if (iFrom < 0)
      break;

    iExt = (pWP->dir[iFrom].btEX & 0x1F) | ((pWP->dir[iFrom].btS2 & 0x7F) << 5);
    iRecs = iExt * RECORDS_PER_EXTENT + pWP->dir[iFrom].btRC;

    iFrom++;
  }

  return iRecs;
}
//-----------------------------------------------------------------------------
int cpmFindFirstFile(WORK_PARAMS *pWP, char *pcaCpmMask)
{
  memcpy(pWP->caCpmFindMask, pcaCpmMask, 11);
  pWP->iCpmFindPosition = 0;

  return cpmFindNextFile(pWP);
}
//-----------------------------------------------------------------------------
int cpmFindNextFile(WORK_PARAMS *pWP)
{
  int ii, jj;
  char *p;

  for (ii = pWP->iCpmFindPosition; ii < pWP->iDirEntries; ii++) {
    if (pWP->dir[ii].cUser == FILL_BYTE || pWP->dir[ii].btEX >= pWP->iExtentsPerDirEntry)
      continue;
    if (pWP->cUsr != -1 && pWP->cUsr != pWP->dir[ii].cUser)
      continue;

    p = pWP->dir[ii].caFN;
    for (jj = 0; jj < 11; jj++) {
      if (pWP->caCpmFindMask[jj] == '?')
        continue;
      if (pWP->caCpmFindMask[jj] != (p[jj] & 0x7F))
        break;
    }
    if (jj == 11) {
      pWP->iCpmFindPosition = ii + 1;
      return ii;
    }
  }

  pWP->iCpmFindPosition = pWP->iDirEntries;

  return -1;
}
//-----------------------------------------------------------------------------
BOOL computeTrackSector(WORK_PARAMS *pWP, int iAU, int iRec, int *pT, int *pS)
{
  int iTotalRecs;

  if (iAU >= pWP->iAllocationUnits || iRec >= pWP->iRecordsPerAU)
    return FALSE;

  iTotalRecs = iAU * pWP->iRecordsPerAU + iRec;
  *pS = iTotalRecs % pWP->iSectorsPerTrack;
  *pT = iTotalRecs / pWP->iSectorsPerTrack + pWP->iSystemTracks;

  return TRUE;
}
//-----------------------------------------------------------------------------
int openCurrentFile(WORK_PARAMS *pWP, BOOL bCreate)
{
  char szFN[MAX_PATH_LEN];
  int iErr;

  strcpy(szFN, pWP->szFilePath);
  strcat(szFN, pWP->szFileName);

  if (bCreate) {
    remove(szFN); // remove first
    pWP->fFile = fopen(szFN, "a+b"); // create if not exist
    if (pWP->fFile != NULL)
      fclose(pWP->fFile);
  }

  if (bCreate)
    pWP->fFile = fopen(szFN, "r+b");
  else
    pWP->fFile = fopen(szFN, "rb");
  if (pWP->fFile == NULL) {
    if (bCreate)
      iErr = printError(ERR_CREATE_FILE, pWP->szFileName, NULL);
    else
      iErr = printError(ERR_OPEN_FILE, pWP->szFileName, NULL);
    return iErr;
  }

  if (bCreate)
    pWP->iFileLength = 0;
  else {
    fseek(pWP->fFile, 0L, SEEK_END);
    pWP->iFileLength = ftell(pWP->fFile);
  }
  fseek(pWP->fFile, 0L, SEEK_SET);

  return ERR_OK;
}
//-----------------------------------------------------------------------------
void closeCurrentFile(WORK_PARAMS *pWP)
{
  if (pWP->fFile != NULL)
    fclose(pWP->fFile);
  pWP->fFile = NULL;
}
//---------------------------------------------------------------------------

