//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "common.h"

//---------------------------------------------------------------------------
int readConfig(WORK_PARAMS *pWP)
{
  char szBuff[1024];
  char *p, *q;
  FILE *fIni;
  int iErr, iFlg, i, j;
  int iSPT, iXLT, iBLS, iEXM, iDSM, iDRM, iOFF, iOFFX, iBSH, iBLM;

  if (*pWP->szImageExt == 0) {
    iErr = printError(ERR_IMG_EXT_MISSING, pWP->szImageName, NULL);
    return iErr;
  }

  strcpy(szBuff, pWP->szExePath);
  strcat(szBuff, pWP->szExeName);
  strcat(szBuff, ".ini");
  fIni = fopen(szBuff, "r");
  if (fIni == NULL) {
    iErr = printError(ERR_CONF_FILE_MISSING, pWP->szExeName, NULL);
    return iErr;
  }

  iSPT = iXLT = iBLS = iEXM = iDSM = iDRM = iOFF = iOFFX = -1;
  iErr = ERR_OK;
  while (TRUE) {
    p = fgets(szBuff, sizeof(szBuff) - 1, fIni);
    if (p == NULL) {  // end of file?
      iErr = printError(ERR_IMG_DEF_MISSING, pWP->szImageExt, NULL); // yes
      break;
    }

    trim(p);
    if (*p != '[')  // begin of section?
      continue;     // no
    p++;
    q = strchr(p, ']'); // check end of section name
    if (q == NULL)
      continue;
    *q++ = 0;
    if (*q != 0)
      continue;
    if (stricmp(pWP->szImageExt, p) != 0) // check extention of image file
      continue;

    iFlg = 0;
    pWP->iSectorsPerTrack = -1;
    while (TRUE) {
      p = fgets(szBuff, sizeof(szBuff) - 1, fIni);
      if (p == NULL)   // end of file?
        break;         // yes
      trim(p);
      if (*p == 0 || *p == ';') // empty line or comment
        continue;
      if (*p == '[')  // next section
        break;
      q = strchr(p, '='); // find value
      if (q == NULL) {
        iErr = printError(ERR_IMG_DEF_ERROR, pWP->szImageExt, NULL); // value missing
        break;
      }

      *q++ = 0;     // ending of key
      trimRight(p); // trim key

      trimLeft(q);  // trim value
      if (*q == 0) {
        iErr = printError(ERR_IMG_DEF_ERROR, pWP->szImageExt, NULL); // value is empty
        break;
      }

      if (strcmp(p, "SPT") == 0) {
        iSPT = parseNumber(q);
        if (iSPT < 1)
          break; // invalid value
        pWP->iSectorsPerTrack = iSPT;
        iFlg |= PAR_SPT;
      }
      else if (strcmp(p, "XLT") == 0) {
        if ((iFlg & PAR_SPT) == 0)
          break; // SPT must precede XLT
        pWP->xlt = (int *) malloc(pWP->iSectorsPerTrack * sizeof(int));
        if (pWP->xlt == NULL) {
          iErr = printError(ERR_OUT_OF_MEMORY, NULL, NULL); // not enough space
          break;
        }

        iXLT = (int) strtol(q, &p, 10);
        if (iXLT == 0) {
          if (*p != 0)
            break;
          for (i = 0; i < pWP->iSectorsPerTrack; i++)
            pWP->xlt[i] = i + 1;
        }
        else {
          for (i = 0; i < pWP->iSectorsPerTrack; i++) {
            pWP->xlt[i] = (int) strtol(q, &p, 10);
            if (pWP->xlt[i] < 1 || pWP->xlt[i] > pWP->iSectorsPerTrack)
              break;
            if (((i + 1) == pWP->iSectorsPerTrack && *p != 0)
                || ((i + 1) < pWP->iSectorsPerTrack && *p != ','))
              break;
            for (j = 0; j < i; j++) {
              if (pWP->xlt[j] == pWP->xlt[i])
                break;
            }
            if (j < i)
              break;
            q = p + 1;
          }
          if (i < pWP->iSectorsPerTrack)
            break; // invalid value
        }
        iFlg |= PAR_XLT;
      }
      else if (strcmp(p, "BLS") == 0) {
        iBLS = parseNumber(q);
        if (iBLS != 1 && iBLS != 2 && iBLS != 4 && iBLS != 8 && iBLS != 16)
          break; // invalid value
        pWP->iAllocationUnitSize = iBLS * 1024;
        iFlg |= PAR_BLS;
      }
      else if (strcmp(p, "EXM") == 0) {
        iEXM = parseNumber(q);
        if (iEXM != 0 && iEXM != 1 && iEXM != 3 && iEXM != 7 && iEXM != 15)
          break; // invalid value
        pWP->iExtentsPerDirEntry = iEXM + 1;
        iFlg |= PAR_EXM;
      }
      else if (strcmp(p, "DSM") == 0) {
        iDSM = parseNumber(q);
        if (iDSM < 1 || iDSM > 65535)
          break; // invalid value
        pWP->iAllocationUnits = iDSM + 1;
        iFlg |= PAR_DSM;
      }
      else if (strcmp(p, "DRM") == 0) {
        iDRM = parseNumber(q);
        if ((iFlg & PAR_BLS) == 0 || iDRM < 0 || iDRM > 65535
            || (((iDRM + 1) * 32) % pWP->iAllocationUnitSize) > 0)
          break; // invalid value
        pWP->iDirEntries = iDRM + 1;
        iFlg |= PAR_DRM;
      }
      else if (strcmp(p, "OFF") == 0) {
        iOFF = parseNumber(q);
        if (iOFF < 0)
          break; // invalid value
        pWP->iSystemTracks = iOFF;
        if ((iFlg & PAR_OFFX) == 0) {
          if (pWP->iSectorsPerTrack < 0)
            break; // invalid value
          pWP->iSystemSize = pWP->iSystemTracks * pWP->iSectorsPerTrack * RECORD_SIZE;
        }
        iFlg |= PAR_OFF;
      }
      else if (strcmp(p, "OFFX") == 0) {
        iOFFX = parseNumber(q);
        if (iOFFX <= 0)
          break; // invalid value
        pWP->iSystemSize = iOFFX * RECORD_SIZE;
        iFlg |= PAR_OFFX;
      }
    }

    break;
  }

  if (fIni != NULL)
    fclose(fIni);

  if (iErr != ERR_OK)
    return iErr;

  if ((iFlg & PAR_ALL) != PAR_ALL)
    iErr = printError(ERR_IMG_DEF_ERROR, pWP->szImageExt, NULL);

  if (iErr != ERR_OK)
    return iErr;

  pWP->iTrackSize = pWP->iSectorsPerTrack * RECORD_SIZE;
  pWP->iTotalDiskSize = pWP->iAllocationUnits * pWP->iAllocationUnitSize;
  pWP->iTotalDiskSize = pWP->iTotalDiskSize + pWP->iTrackSize - 1;
  pWP->iTotalDiskSize = (pWP->iTotalDiskSize / pWP->iTrackSize) * pWP->iTrackSize;
  pWP->iTotalDiskSize = pWP->iTotalDiskSize + pWP->iSystemSize;

  if (pWP->iTotalDiskSize > MAX_CPM_DISK_SIZE)
    iErr = printError(ERR_IMG_DEF_ERROR, pWP->szImageExt, NULL);  // image is too large - bad definition
  else {
    pWP->dir = (FCB *) malloc(pWP->iDirEntries * sizeof(FCB));
    pWP->iAllocVectorSize = (pWP->iAllocationUnits + 7) / 8;
    pWP->au = (BYTE *) malloc(pWP->iAllocVectorSize);
    pWP->iRecordsPerAU = pWP->iAllocationUnitSize / RECORD_SIZE;
    pWP->iAuPerDirEntry = (pWP->iAllocationUnits > 256) ? 8 : 16;
  }

  if (iErr != ERR_OK)
    return iErr;

  if (!pWP->bImageTypeInfo)
    return iErr;

  iBSH = 0;
  iBLM = 0;
  switch (iBLS) {
    case 1 :
      iBSH = 3;
      iBLM = 7;
      break;
    case 2 :
      iBSH = 4;
      iBLM = 15;
      break;
    case 4 :
      iBSH = 5;
      iBLM = 31;
      break;
    case 8 :
      iBSH = 6;
      iBLM = 63;
      break;
    case 16 :
      iBSH = 7;
      iBLM = 127;
      break;
  }

  // iSPT, iXLT, iBLS, iEXM, iDSM, iDRM, iOFF, iOFFX
  printf("Definition parameters for `%s`\n", pWP->szImageExt);
  printf("  SPT=%d\n", iSPT);
  printf("  XLT=");
  if (iXLT == 0)
    printf("0");
  else {
    for (i = 0; i < pWP->iSectorsPerTrack; i++) {
      if (i > 0)
        printf(",");
      printf("%d", pWP->xlt[i]);
    }
  }
  printf("\n");
  printf("  BLS=%d (BSH=%d, BLM=%d)\n", iBLS, iBSH, iBLM);
  printf("  EXM=%d\n", iEXM);
  printf("  DSM=%d\n", iDSM);
  printf("  DRM=%d\n", iDRM);
  printf("  OFF=%d\n", iOFF);
  if (iOFFX > 0)
    printf("  OFFX=%d\n", iOFFX);
  printf("\n");

  printf("Other info:\n", pWP->szImageExt);
  printf("  Sectors Per Track: %d\n", pWP->iSectorsPerTrack);
  printf("  Allocation Unit Size: %d Bytes\n", pWP->iAllocationUnitSize);
  printf("  Extents Per Dir Entry: %d\n", pWP->iExtentsPerDirEntry);
  printf("  Allocation Units: %d\n", pWP->iAllocationUnits);
  printf("  Directory Entries: %d\n", pWP->iDirEntries);
  printf("  System Size: %d Bytes\n", pWP->iSystemSize);
  printf("  Track Size: %d Bytes\n", pWP->iTrackSize);
  printf("  Total Disk Size: %d Bytes (%g kB)\n", pWP->iTotalDiskSize, pWP->iTotalDiskSize / 1024.0);

  return iErr;
}
//-----------------------------------------------------------------------------
int openImage(WORK_PARAMS *pWP)
{
  char szFN[MAX_PATH_LEN];
  int ii, iErr;
  BOOL bCanCreate, bReadOnly;

  strcpy(szFN, pWP->szImagePath);
  strcat(szFN, pWP->szImageName);

  bCanCreate = (pWP->bFormatFull || pWP->bFormatQuick || pWP->bAddFiles);
  bReadOnly = (pWP->bCatalogue || pWP->bDumpAscii || pWP->bDumpHex
               || pWP->bExtractFiles || pWP->bExtractSystem);

  if (bCanCreate) {
    if (pWP->bFormatFull)
      pWP->fImage = fopen(szFN, "w+b"); // create or overwrite if exist
    else
      pWP->fImage = fopen(szFN, "a+b"); // create if not exist
    if (pWP->fImage != NULL)
      fclose(pWP->fImage);
  }

  if (bReadOnly)
    pWP->fImage = fopen(szFN, "rb");
  else
    pWP->fImage = fopen(szFN, "r+b");
  if (pWP->fImage == NULL) {
    if (bCanCreate)
      iErr = printError(ERR_CREATE_IMAGE, pWP->szImageName, NULL);
    else
      iErr = printError(ERR_OPEN_IMAGE, pWP->szImageName, NULL);
    return iErr;
  }

  fseek(pWP->fImage, 0L, SEEK_END);
  if (bCanCreate && ftell(pWP->fImage) == 0) {
    pWP->bFormatFull = TRUE;
    memset(pWP->pbtBuff, FILL_BYTE, RECORD_SIZE);

    fseek(pWP->fImage, 0L, SEEK_SET);
    for (ii = 0; ii < pWP->iTotalDiskSize; ii += RECORD_SIZE) {
      if (fwrite(pWP->pbtBuff, RECORD_SIZE, 1, pWP->fImage) != 1)
        return printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
    }
    printf("Image '%s' was created.\n", pWP->szImageName);
  }

  if (ftell(pWP->fImage) != pWP->iTotalDiskSize)
    return printError(ERR_IMAGE_SIZE, pWP->szImageName, NULL);

  return ERR_OK;
}
//-----------------------------------------------------------------------------
void closeImage(WORK_PARAMS *pWP)
{
  if (pWP->fImage != NULL)
    fclose(pWP->fImage);
  pWP->fImage = NULL;
}
//-----------------------------------------------------------------------------
int writeBoot(WORK_PARAMS *pWP, BOOL bWriteTail)
{
  if (pWP->iSystemSize > 0) {
    memset(pWP->pbtBuff, FILL_BYTE, RECORD_SIZE);
    *pWP->pbtBuff = 0xC9; // instruction RET

    fseek(pWP->fImage, 0L, SEEK_SET);
    if (fwrite(pWP->pbtBuff, RECORD_SIZE, 1, pWP->fImage) != 1)
      return printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
  }

  if (bWriteTail) {
    memset(pWP->pbtBuff, 0, RECORD_SIZE);
    sprintf((char *) pWP->pbtBuff, "This disk was formatted with '%s v%d.%d' programmed by %s",
            PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);

    fseek(pWP->fImage, pWP->iTotalDiskSize - RECORD_SIZE, SEEK_SET);
    if (fwrite(pWP->pbtBuff, RECORD_SIZE, 1, pWP->fImage) != 1)
      return printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
  }

  if (pWP->bFormatFull)
    printf("Image '%s' was formated.\n", pWP->szImageName);

  return ERR_OK;
}
//---------------------------------------------------------------------------
int readSector(WORK_PARAMS *pWP, BYTE *pbtBuf, int iTrack, int iSector)
{
  int iSeek = pWP->iSystemSize
            + (iTrack - pWP->iSystemTracks) * pWP->iTrackSize
            + ((pWP->xlt[iSector] - 1) * RECORD_SIZE);
  fseek(pWP->fImage, iSeek, SEEK_SET);

  if (fread(pbtBuf, RECORD_SIZE, 1, pWP->fImage) == 1)
    return ERR_OK;
  return printError(ERR_READ_IMAGE, pWP->szImageName, NULL);
}
//---------------------------------------------------------------------------
int writeSector(WORK_PARAMS *pWP, BYTE *pbtBuf, int iTrack, int iSector)
{
  int iSeek = pWP->iSystemSize
            + (iTrack - pWP->iSystemTracks) * pWP->iTrackSize
            + ((pWP->xlt[iSector] - 1) * RECORD_SIZE);
  fseek(pWP->fImage, iSeek, SEEK_SET);

  if (fwrite(pbtBuf, RECORD_SIZE, 1, pWP->fImage) == 1)
    return ERR_OK;
  return printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
}
//-----------------------------------------------------------------------------
int readDirectory(WORK_PARAMS *pWP)
{
  int ii, jj, iT, iS;
  int iDirAU, iAUnum, iErr;
  char str[64];
  FCB *pFCB;
  BOOL bWasError = FALSE;

  iDirAU = (pWP->iDirEntries * DIR_ENTRY_SIZE) / pWP->iAllocationUnitSize;
  for (ii = 0; ii < iDirAU; ii++) {
    for (jj = 0; jj < pWP->iRecordsPerAU; jj++) {
      if (!computeTrackSector(pWP, ii, jj, &iT, &iS))
        return printError(ERR_FATAL_ERROR, NULL, NULL);
      iErr = readSector(pWP, (BYTE *) &pWP->dir[ii * pWP->iRecordsPerAU * DIR_ENTRIES_PER_RECORD + jj * DIR_ENTRIES_PER_RECORD], iT, iS);
      if (iErr != ERR_OK)
        return iErr;
    }
  }

  // unmark all allocation units
  for (ii = 0; ii < pWP->iAllocVectorSize; ii++)
    pWP->au[ii] = 0;

  // mark allocation units occupied by directory itself
  for (ii = 0; ii < iDirAU; ii++)
    markAU(pWP, ii, TRUE);

  // mark allocation units occupied by files
  for (ii = 0; ii < pWP->iDirEntries; ii++) {
    pFCB = (FCB *) &pWP->dir[ii];
    if (pFCB->cUser == FILL_BYTE) // empty directory entry ?
      continue;

    prepareNameAttr(pWP, ii);
    if (pFCB->cUser < 0 || pFCB->cUser > 15) {
      sprintf(str, "invalid user number: %d", (int) pFCB->cUser);
      iErr = printError(ERR_DIR_ENTRY, pWP->fna.szFN, str);
      bWasError = TRUE;
      if (!pWP->bIgnoreInvalidDE)
        break;
    }

    if (!checkFileName(pWP, NULL, pWP->fna.szFN)) {
      iErr = printError(ERR_DIR_ENTRY, pWP->fna.szFN, "bad file name");
      bWasError = TRUE;
      if (!pWP->bIgnoreInvalidDE)
        break;
    }

    if (pWP->iAuPerDirEntry == 8) {
      // allocation units numbers are WORDs
      for (jj = 0; jj < 16; jj += 2) {
        iAUnum = (int) *(WORD *) (&pFCB->btaAU[jj]);
        if (iAUnum == 0)
          break;
        if (!markAU(pWP, iAUnum, TRUE)) {
          sprintf(str, "invalid AU number: %d", iAUnum);
          iErr = printError(ERR_DIR_ENTRY, pWP->fna.szFN, str);
          bWasError = TRUE;
          if (!pWP->bIgnoreInvalidDE)
            break;
        }
      }
    }
    else {
      // allocation units numbers are BYTEs
      for (jj = 0; jj < 16; jj++) {
        iAUnum = (int) pFCB->btaAU[jj];
        if (iAUnum == 0)
          break;
        if (!markAU(pWP, iAUnum, TRUE)) {
          sprintf(str, "invalid AU number: %d", iAUnum);
          iErr = printError(ERR_DIR_ENTRY, pWP->fna.szFN, str);
          bWasError = TRUE;
          if (!pWP->bIgnoreInvalidDE)
            return iErr;
        }
      }
    }
    if (iErr != ERR_OK && !pWP->bIgnoreInvalidDE)
      break;
  }

  if (bWasError && pWP->bIgnoreInvalidDE)
    printf("\n");

  return ERR_OK;
}
//-----------------------------------------------------------------------------
int writeDirectory(WORK_PARAMS *pWP)
{
  int ii, jj, iT, iS, iErr;
  int iDirAU = (pWP->iDirEntries * DIR_ENTRY_SIZE) / pWP->iAllocationUnitSize;

  for (ii = 0; ii < iDirAU; ii++) {
    for (jj = 0; jj < pWP->iRecordsPerAU; jj++) {
      if (!computeTrackSector(pWP, ii, jj, &iT, &iS))
        return printError(ERR_FATAL_ERROR, NULL, NULL);
      iErr = writeSector(pWP, (BYTE *) &pWP->dir[ii * pWP->iRecordsPerAU * DIR_ENTRIES_PER_RECORD + jj * DIR_ENTRIES_PER_RECORD], iT, iS);
      if (iErr != ERR_OK)
        return iErr;
    }
  }

  return ERR_OK;
}
//---------------------------------------------------------------------------
int cleanDirectory(WORK_PARAMS *pWP)
{
  int ii;
  int iDirSize = pWP->iDirEntries * DIR_ENTRY_SIZE;

  fseek(pWP->fImage, pWP->iSystemSize, SEEK_SET);

  memset(pWP->pbtBuff, FILL_BYTE, RECORD_SIZE);

  for (ii = 0; ii < iDirSize; ii += RECORD_SIZE) {
    if (fwrite(pWP->pbtBuff, RECORD_SIZE, 1, pWP->fImage) != 1)
      return printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
  }

  if (pWP->bFormatQuick)
    printf("Directory of image '%s' was cleaned.\n", pWP->szImageName);

  return ERR_OK;
}
//---------------------------------------------------------------------------

