//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "common.h"

#if WIN
  #include <dir.h>
#else
  #include <dirent.h>
#endif

//-----------------------------------------------------------------------------
void catalogue(WORK_PARAMS *pWP)
{
  int ii, jj, iFiles, iTotal, iDirEntryNum;
  int *piaSort;
  FILE_NAME_ATTR *pfnaDirfiles;

  if (!checkFileName(pWP, pWP->caCpmFileMask, pWP->szFileMask)) {
    printError(ERR_BAD_FILE_NAME, pWP->szFileMask, NULL);
    return;
  }

  piaSort = (int *) malloc(pWP->iDirEntries * sizeof(int));
  pfnaDirfiles = (FILE_NAME_ATTR *) malloc(pWP->iDirEntries * sizeof(FILE_NAME_ATTR));

  iFiles = 0;
  iTotal = 0;

  iDirEntryNum = cpmFindFirstFile(pWP, pWP->caCpmFileMask);
  while (iDirEntryNum >= 0) {
    prepareNameAttr(pWP, iDirEntryNum);
    pWP->fna.iTotalRecs = calcFileRecs(pWP, &pWP->dir[iDirEntryNum].cUser);
    pWP->fna.iSize = (pWP->fna.iTotalRecs + pWP->iRecordsPerAU - 1)
                   / pWP->iRecordsPerAU * (pWP->iAllocationUnitSize / 1024);
    memcpy((char *) &pfnaDirfiles[iFiles], (char *) &pWP->fna, sizeof(FILE_NAME_ATTR));

    for (jj = 0; jj < iFiles; jj++) {
      if (strcmp(pWP->fna.szFN, pfnaDirfiles[piaSort[jj]].szFN) < 0)
        break;
    }
    for (ii = iFiles; ii > jj; ii--)
      piaSort[ii] = piaSort[ii - 1];
    piaSort[jj] = iFiles;

    iFiles++;
    iTotal += pWP->fna.iSize;

    iDirEntryNum = cpmFindNextFile(pWP);
  }

  printf("Catalogue of '%s', mask: '%s'\n\n", pWP->szImageName, pWP->szFileMask);
  printf(" U Name          RECS  SIZE  RH\t U Name          RECS  SIZE  RH\n");
  printf("-------------------------------\t-------------------------------\n");

  for (ii = 0; ii < iFiles; ii++) {
    jj = piaSort[ii];

    while (strlen(pfnaDirfiles[jj].szFN) < 12)
      strcat(pfnaDirfiles[jj].szFN, " ");

    printf("%2d %s  %4d %4dK  %c%c",
           pfnaDirfiles[jj].cUser, pfnaDirfiles[jj].szFN,
           pfnaDirfiles[jj].iTotalRecs, pfnaDirfiles[jj].iSize,
           (pfnaDirfiles[jj].bReadOnly) ? 'R' : '-',
           (pfnaDirfiles[jj].bHidden) ? 'H' : '-');
    if (ii % 2)
      printf("\n");
    else
      printf("\t");

    if (((ii + 1) % 32) == 0) {
      pressEnter("Press Enter to continue...");
      printf("\r");
    }
  }

  if (iFiles % 2)
     printf("\n");
  printf("\n");

  printf("%dK in %d file(s), %dK free\n", iTotal, iFiles, getFreeAuCount(pWP) * (pWP->iAllocationUnitSize / 1024));
  printf("%d directory entries free\n\n", getFreeDirEntriesCount(pWP));

  free(piaSort);
  free(pfnaDirfiles);
}
//-----------------------------------------------------------------------------
void addFiles(WORK_PARAMS *pWP)
{
  int iDirEntryNum, iAddedFilesCount, iAuCnt, iTotalRecs;
  int jj, iT, iS, iLen, iAUnum;
  char szFN[MAX_PATH_LEN];
#if WIN
  struct ffblk ffd;
#else
  struct dirent *pDirEnt;
  DIR *pDir;
  char caMask[11];
#endif

  strcpy(szFN, pWP->szFilePath);
  strcat(szFN, pWP->szFileMask);

  printf("Adding file(s) to '%s'\n  path: '%s', mask: '%s'\n", pWP->szImageName, pWP->szFilePath, pWP->szFileMask);
  if (*pWP->szFileNameCpm)
    printf("  as '%s'\n", pWP->szFileNameCpm);
  printf("\n");

#if WIN
  if (findfirst(szFN, &ffd, FA_RDONLY | FA_ARCH)) {
    if (_doserrno == ENOENT)
      printf("No matching files found.\n\n", szFN);
    else
      printf("Path or file/mask `%s` is invalid!!!\n\n", szFN);
    return;
  }
#else
  pDir = opendir(pWP->szFilePath);
  if (pDir == NULL || !checkFileName(pWP, caMask, pWP->szFileMask)) {
    printf("Path or file/mask `%s` is invalid!!!\n\n", szFN);
    return;
  }
#endif

  iDirEntryNum = 0;
  iAddedFilesCount = 0;

#if WIN
  do {
    strcpy(pWP->szFileName, ffd.ff_name);
    strupr(pWP->szFileName);
#else
  while ((pDirEnt = readdir(pDir)) != NULL) {
    if (pDirEnt->d_type != DT_REG && pDirEnt->d_type != DT_LNK)
      continue;
    strcpy(pWP->szFileName, pDirEnt->d_name);
#endif

    if (*pWP->szFileNameCpm) {
      if (!checkFileName(pWP, pWP->caCpmFileMask, pWP->szFileNameCpm)) {
        printError(ERR_BAD_FILE_NAME, pWP->szFileNameCpm, NULL);
        break;
      }

      if (strchr(pWP->caCpmFileMask, '?') != NULL) {
        printf("File name '%s' must be unambiguous!\n\n", pWP->szFileNameCpm);
        break;
      }

      if (pWP->bOverwrite)
        eraseFiles(pWP, TRUE);
    }
    else if (!checkFileName(pWP, pWP->caCpmFileMask, pWP->szFileName)) {
      printError(ERR_BAD_FILE_NAME, pWP->szFileName, NULL);
      continue;
    }

#if LINUX
    for (jj = 0; jj < 11; jj++) {
      if (caMask[jj] == '?')
        continue;
      if (caMask[jj] != pWP->caCpmFileMask[jj]) {
        jj = -1;
        break;
      }
    }
    if (jj < 0)
      continue;
#endif

    // check for existence of the file with this name on the image
    if (cpmFindFirstFile(pWP, pWP->caCpmFileMask) >= 0) {
      if (*pWP->szFileNameCpm)
        printError(ERR_FILE_EXIST, pWP->szFileNameCpm, NULL);
      else
        printError(ERR_FILE_EXIST, pWP->szFileName, NULL);
      continue;
    }

    // open file that should be added to the image
    if (openCurrentFile(pWP, FALSE) != ERR_OK) {
      closeCurrentFile(pWP);
      continue;
    }

    // check file size
    if (pWP->iFileLength > MAX_CPM_FILE_SIZE) {
      printError(ERR_FILE_TOO_LONG, pWP->szFileName, &pWP->iFileLength);
      closeCurrentFile(pWP);
      continue;
    }

    // check number of directory entries that will be occupied by file
    if (((pWP->iFileLength + EXTENT_SIZE * pWP->iExtentsPerDirEntry - 1) / (EXTENT_SIZE * pWP->iExtentsPerDirEntry)) > getFreeDirEntriesCount(pWP)) {
      printError(ERR_DIRECTORY_FULL, NULL, NULL);
      break;
    }

    // check number of allocation units that file needs
    if (((pWP->iFileLength + pWP->iAllocationUnitSize - 1) / pWP->iAllocationUnitSize) > getFreeAuCount(pWP)) {
      printError(ERR_DISK_FULL, NULL, NULL);
      break;
    }

    // compute total records
    iTotalRecs = (pWP->iFileLength + RECORD_SIZE - 1) / RECORD_SIZE;

    iAuCnt = 0;
    do {
      if ((iAuCnt % pWP->iAuPerDirEntry) == 0) {
        if (iAuCnt > 0) {
          jj = (iAuCnt - 1) / (pWP->iAuPerDirEntry / pWP->iExtentsPerDirEntry);
          pWP->dir[iDirEntryNum].btEX = (jj & 0x1F);
          if (jj > 0x1F)
            pWP->dir[iDirEntryNum].btS2 = ((jj >> 5) & 0x7F);
          pWP->dir[iDirEntryNum].btRC = 128;
        }
        iDirEntryNum = findFreeDirEntry(pWP);
        memset(&pWP->dir[iDirEntryNum], 0, sizeof(FCB));
        if (pWP->cUsr == -1)
          pWP->dir[iDirEntryNum].cUser = 0;
        else
          pWP->dir[iDirEntryNum].cUser = pWP->cUsr;
        memcpy(pWP->dir[iDirEntryNum].caFN, pWP->caCpmFileMask, 11);
      }

      if (pWP->iFileLength == 0) // for zero file length
        break;

      iAUnum = findFirstFreeAU(pWP);
      markAU(pWP, iAUnum, TRUE);
      if (pWP->iAuPerDirEntry == 16)
        pWP->dir[iDirEntryNum].btaAU[iAuCnt % 16] = (BYTE) iAUnum;
      else
        *(WORD *) &pWP->dir[iDirEntryNum].btaAU[(iAuCnt % 8) * 2] = (WORD) iAUnum;

      memset(pWP->pbtBuff, FILL_BYTE, BUFFER_SIZE);
      iLen = (pWP->iFileLength > pWP->iAllocationUnitSize) ? pWP->iAllocationUnitSize : pWP->iFileLength;
      if (fread(pWP->pbtBuff, iLen, 1, pWP->fFile) != 1) {
        printError(ERR_READ_FILE, pWP->szFileName, NULL);
        return;
      }

      for (jj = 0; jj < pWP->iRecordsPerAU; jj++) {
        if (!computeTrackSector(pWP, iAUnum, jj, &iT, &iS)) {
          printError(ERR_FATAL_ERROR, NULL, NULL);
          return;
        }
        if (writeSector(pWP, pWP->pbtBuff + jj * RECORD_SIZE, iT, iS) != ERR_OK) {
          printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
          return;
        }
      }

      iAuCnt++;
      pWP->iFileLength -= iLen;

    } while (pWP->iFileLength > 0);

    if (iAuCnt > 0) {  // close last extent of the file with no zero length
      jj = (iAuCnt - 1) / (pWP->iAuPerDirEntry / pWP->iExtentsPerDirEntry);
      pWP->dir[iDirEntryNum].btEX = (jj & 0x1F);
      if (jj > 0x1F)
        pWP->dir[iDirEntryNum].btS2 = ((jj >> 5) & 0x7F);
      pWP->dir[iDirEntryNum].btRC = iTotalRecs % 128;
      if (pWP->dir[iDirEntryNum].btRC == 0 && iTotalRecs > 0)
        pWP->dir[iDirEntryNum].btRC = 128;
    }

    closeCurrentFile(pWP);
    iAddedFilesCount++;

    if (*pWP->szFileNameCpm)
      printf("File '%s' added\n", pWP->szFileNameCpm);
    else
      printf("File '%s' added\n", pWP->szFileName);

#if LINUX
  }
  closedir(pDir);
#else
  } while (findnext(&ffd) == 0);
  findclose(&ffd);
#endif

  if (writeDirectory(pWP) != ERR_OK)
    printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);

  printf("\nDone. %d file(s) added.\n\n", iAddedFilesCount);
}
//-----------------------------------------------------------------------------
void eraseFiles(WORK_PARAMS *pWP, BOOL bForSingleAdd)
{
  char szEntry[13];
  int iDirEntryNum, iDirEntryNum2, iErasedFiles;
  int ii, iAUnum;

  if (!bForSingleAdd) {
    if (!checkFileName(pWP, pWP->caCpmFileMask, pWP->szFileMask)) {
      printError(ERR_BAD_FILE_NAME, pWP->szFileMask, NULL);
      return;
    }
  }

  if (!pWP->bOverwrite)
    printf("Erasing file(s) from '%s', mask: '%s'\n\n", pWP->szImageName, pWP->szFileMask);

  iErasedFiles = 0;
  iDirEntryNum = cpmFindFirstFile(pWP, pWP->caCpmFileMask);
  while (iDirEntryNum >= 0) {
    memcpy(szEntry, &pWP->dir[iDirEntryNum].cUser, 12);
    iDirEntryNum2 = iDirEntryNum;
    do {
      // free Dir Entry
      pWP->dir[iDirEntryNum2].cUser = FILL_BYTE;
      // free Allocation Units
      for (ii = 0; ii < pWP->iAuPerDirEntry; ii++) {
        if (pWP->iAuPerDirEntry == 16)
          iAUnum = pWP->dir[iDirEntryNum2].btaAU[ii];
        else
          iAUnum = *(WORD *) &pWP->dir[iDirEntryNum].btaAU[ii * 2];
        if (iAUnum > 0)
          markAU(pWP, iAUnum, FALSE);
      }
      // next Dir Entry for this file
      iDirEntryNum2 = iDirEntryNum + 1;
    } while ((iDirEntryNum2 = findDirEntry(pWP, szEntry, iDirEntryNum2)) >= 0);
    iErasedFiles++;

    memcpy(szEntry, &pWP->dir[iDirEntryNum].caFN, 8);
    szEntry[8] = 0;
    trimRight(szEntry);
    strcat(szEntry, ".");
    iDirEntryNum2 = strlen(szEntry);
    memcpy(szEntry + iDirEntryNum2, &pWP->dir[iDirEntryNum].caFT, 3);
    szEntry[iDirEntryNum2 + 3] = 0;
    trimRight(szEntry);
    if (!pWP->bOverwrite)
      printf("File '%s' eresed\n", szEntry);

    iDirEntryNum = cpmFindNextFile(pWP);
  }

  if (!bForSingleAdd) {
    if (writeDirectory(pWP) != ERR_OK)
      printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
  }

  if (!pWP->bOverwrite)
    printf("\nDone. %d file(s) erased.\n\n", iErasedFiles);
}
//-----------------------------------------------------------------------------
void extractFiles(WORK_PARAMS *pWP)
{
  char caEntry[12], c;
  int iDirEntryNum, iDirEntryNum2, iCnt, ii, jj, iAU, iT, iS;

  if (!checkFileName(pWP, pWP->caCpmFileMask, pWP->szFileMask)) {
    printError(ERR_BAD_FILE_NAME, pWP->szFileMask, NULL);
    return;
  }

  printf("Extracting file(s) from '%s'\n  path: '%s', mask: '%s'\n\n", pWP->szImageName, pWP->szFilePath, pWP->szFileMask);

  iCnt = 0;
  iDirEntryNum = cpmFindFirstFile(pWP, pWP->caCpmFileMask);
  while (iDirEntryNum >= 0) {
    memcpy(caEntry, &pWP->dir[iDirEntryNum].cUser, 12);
    prepareNameAttr(pWP, iDirEntryNum);
    pWP->fna.iTotalRecs = calcFileRecs(pWP, caEntry);
    iDirEntryNum2 = iDirEntryNum;

    strcpy(pWP->szFileName, pWP->fna.szFN);
    for (ii = strlen(pWP->szFileName) - 1; ii >= 0; ii--) {
      c = pWP->szFileName[ii];
      if (c == '\\' || c == '/' || c == '"' || c == '|')
        pWP->szFileName[ii] = '_';
    }
    if (openCurrentFile(pWP, TRUE) != ERR_OK)
      return;

    do {
      for (ii = 0; ii < 16; ii++) {
        if (pWP->iAuPerDirEntry == 8)
          iAU = *(WORD *) &pWP->dir[iDirEntryNum2].btaAU[ii++];
        else
          iAU = pWP->dir[iDirEntryNum2].btaAU[ii];
        if (iAU == 0)
          break;

        for (jj = 0; jj < pWP->iRecordsPerAU && pWP->fna.iTotalRecs > 0; jj++) {
          if (!computeTrackSector(pWP, iAU, jj, &iT, &iS)) {
            printError(ERR_FATAL_ERROR, NULL, NULL);
            return;
          }
          if (readSector(pWP, pWP->pbtBuff, iT, iS) != ERR_OK) {
            printError(ERR_READ_IMAGE, pWP->szImageName, NULL);
            return;
          }
          if (fwrite(pWP->pbtBuff, RECORD_SIZE, 1, pWP->fFile) != 1) {
            printError(ERR_WRITE_FILE, pWP->szFileName, NULL);
            return;
          }
          pWP->fna.iTotalRecs--;
        }
      }
    } while ((iDirEntryNum2 = findDirEntry(pWP, caEntry, ++iDirEntryNum2)) >= 0);

    closeCurrentFile(pWP);
    iCnt++;

    if (strcmp(pWP->szFileName, pWP->fna.szFN) == 0)
      printf("File '%s' extracted\n", pWP->szFileName);
    else
      printf("File '%s' extracted as %s\n", pWP->fna.szFN, pWP->szFileName);

    iDirEntryNum = cpmFindNextFile(pWP);
  }

  if (iCnt == 0 && !pWP->bHasWildcard)
    printError(ERR_FILE_NOT_EXIST, pWP->szFileMask, NULL);
  else
    printf("\nDone. %d file(s) extracted.\n\n", iCnt);
}
//-----------------------------------------------------------------------------
void asciiDump(WORK_PARAMS *pWP)
{
  char caEntry[12], c;
  int iDirEntryNum, ii, jj, kk, iAU, iT, iS, iLines;

  if (!checkFileName(pWP, pWP->caCpmFileMask, pWP->szFileMask)) {
    printError(ERR_BAD_FILE_NAME, pWP->szFileMask, NULL);
    return;
  }

  if (strchr(pWP->caCpmFileMask, '?') != NULL) {
    printf("File name '%s' must be unambiguous!\n\n", pWP->szFileMask);
    return;
  }

  printf("Content of the file '%s' from '%s'\n\n", pWP->szFileMask, pWP->szImageName);

  iDirEntryNum = cpmFindFirstFile(pWP, pWP->caCpmFileMask);
  if (iDirEntryNum < 0) {
    printf("File '%s' not found\n\n", pWP->szFileMask);
    return;
  }
  else {
    memcpy(caEntry, &pWP->dir[iDirEntryNum].cUser, 12);
    prepareNameAttr(pWP, iDirEntryNum);
    pWP->fna.iTotalRecs = calcFileRecs(pWP, caEntry);

    iLines = 0;
    do {
      for (ii = 0; ii < 16; ii++) {
        if (pWP->iAuPerDirEntry == 8)
          iAU = *(WORD *) &pWP->dir[iDirEntryNum].btaAU[ii++];
        else
          iAU = pWP->dir[iDirEntryNum].btaAU[ii];
        if (iAU == 0)
          break;

        for (jj = 0; jj < pWP->iRecordsPerAU && pWP->fna.iTotalRecs > 0; jj++) {
          if (!computeTrackSector(pWP, iAU, jj, &iT, &iS)) {
            printError(ERR_FATAL_ERROR, NULL, NULL);
            return;
          }
          if (readSector(pWP, pWP->pbtBuff, iT, iS) != ERR_OK) {
            printError(ERR_READ_IMAGE, pWP->szImageName, NULL);
            return;
          }

          for (kk = 0; kk < RECORD_SIZE; kk++) {
            c = (char) pWP->pbtBuff[kk];
            if (c == 0x1A)  // EOF
              break;
            printf("%c", c);

            if (c == '\n') {
              iLines++;
              if (iLines == LINES_PER_DUMP_PAGE) {
                pressEnter(" ... Press Enter to continue ...");
                iLines = 0;
              }
            }
          }

          pWP->fna.iTotalRecs--;
        }
      }
    } while ((iDirEntryNum = findDirEntry(pWP, caEntry, ++iDirEntryNum)) >= 0);

    printf("\n\n");
  }
}
//-----------------------------------------------------------------------------
void hexDump(WORK_PARAMS *pWP)
{
  char caEntry[12], c;
  int iDirEntryNum, ii, jj, kk, ll, iAU, iT, iS, iAdr, iLines;

  if (!checkFileName(pWP, pWP->caCpmFileMask, pWP->szFileMask)) {
    printError(ERR_BAD_FILE_NAME, pWP->szFileMask, NULL);
    return;
  }

  if (strchr(pWP->caCpmFileMask, '?') != NULL) {
    printf("File name '%s' must be unambiguous!\n\n", pWP->szFileMask);
    return;
  }

  printf("Hex Dump of the file '%s' from '%s'\n\n", pWP->szFileMask, pWP->szImageName);

  iDirEntryNum = cpmFindFirstFile(pWP, pWP->caCpmFileMask);
  if (iDirEntryNum < 0) {
    printf("File '%s' not found\n\n", pWP->szFileMask);
    return;
  }
  else {
    memcpy(caEntry, &pWP->dir[iDirEntryNum].cUser, 12);
    prepareNameAttr(pWP, iDirEntryNum);
    pWP->fna.iTotalRecs = calcFileRecs(pWP, caEntry);

    iLines = 0;
    iAdr = 0;
    do {
      for (ii = 0; ii < 16; ii++) {
        if (pWP->iAuPerDirEntry == 8)
          iAU = *(WORD *) &pWP->dir[iDirEntryNum].btaAU[ii++];
        else
          iAU = pWP->dir[iDirEntryNum].btaAU[ii];
        if (iAU == 0)
          break;

        for (jj = 0; jj < pWP->iRecordsPerAU && pWP->fna.iTotalRecs > 0; jj++) {
          if (!computeTrackSector(pWP, iAU, jj, &iT, &iS)) {
            printError(ERR_FATAL_ERROR, NULL, NULL);
            return;
          }
          if (readSector(pWP, pWP->pbtBuff, iT, iS) != ERR_OK) {
            printError(ERR_READ_IMAGE, pWP->szImageName, NULL);
            return;
          }

          for (kk = 0; kk < RECORD_SIZE; kk += 16) {
            printf("%06X  ", iAdr);
            for (ll = 0; ll < 16; ll++)
              printf("%02X ", pWP->pbtBuff[kk + ll]);
            printf(" ");
            for (ll = 0; ll < 16; ll++) {
              c = pWP->pbtBuff[kk + ll];
              if (c < 32 || c == 127)
                c = '.';
              printf("%c", c);
            }
            printf("\n");

            iAdr += 16;
            iLines++;
            if (iLines == LINES_PER_DUMP_PAGE) {
              pressEnter(" ... Press Enter to continue ...");
              iLines = 0;
            }
          }

          pWP->fna.iTotalRecs--;
        }
      }
    } while ((iDirEntryNum = findDirEntry(pWP, caEntry, ++iDirEntryNum)) >= 0);

    printf("\n\n");
  }
}
//-----------------------------------------------------------------------------
void writeSys(WORK_PARAMS *pWP)
{
  int iLen;

  if (pWP->iSystemSize == 0) {
    printError(ERR_IMG_NO_SYS, pWP->szImageName, NULL);
    return;
  }

  strcpy(pWP->szFileName, pWP->szFileMask);
  if (openCurrentFile(pWP, FALSE) != ERR_OK)
    return;

  if (pWP->iFileLength > pWP->iSystemSize) {
    closeCurrentFile(pWP);
    printError(ERR_IMG_SYS_TOO_LONG, pWP->szFileName, &pWP->iSystemSize);
    return;
  }

  fseek(pWP->fFile, 0L, SEEK_SET);
  fseek(pWP->fImage, 0L, SEEK_SET);
  while (pWP->iFileLength > 0) {
    iLen = (pWP->iFileLength > BUFFER_SIZE) ? BUFFER_SIZE : pWP->iFileLength;
    if (fread(pWP->pbtBuff, iLen, 1, pWP->fFile) != 1) {
      printError(ERR_READ_FILE, pWP->szFileName, NULL);
      return;
    }

    if (fwrite(pWP->pbtBuff, iLen, 1, pWP->fImage) != 1) {
      printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
      return;
    }
    pWP->iFileLength -= iLen;
  }

  closeCurrentFile(pWP);

  printf("\nDone. System written.\n\n");
}
//-----------------------------------------------------------------------------
void copySys(WORK_PARAMS *pWP)
{
  int iLen;
  char *p;

  if (pWP->iSystemSize == 0) {
    printError(ERR_IMG_NO_SYS, pWP->szImageName, NULL);
    return;
  }

  strcpy(pWP->szFileName, pWP->szFileMask);
  p = strrchr(pWP->szFileName, '.');
  if (p == NULL || strcmp(p + 1, pWP->szImageExt) != 0) {
    closeCurrentFile(pWP);
    printError(ERR_IMG_BAD_TYPE, pWP->szFileName, pWP->szImageName);
    return;
  }

  if (openCurrentFile(pWP, FALSE) != ERR_OK)
    return;

  fseek(pWP->fFile, 0L, SEEK_SET);
  fseek(pWP->fImage, 0L, SEEK_SET);
  while (pWP->iSystemSize > 0) {
    iLen = (pWP->iSystemSize > BUFFER_SIZE) ? BUFFER_SIZE : pWP->iSystemSize;
    if (fread(pWP->pbtBuff, iLen, 1, pWP->fFile) != 1) {
      printError(ERR_READ_FILE, pWP->szFileName, NULL);
      return;
    }

    if (fwrite(pWP->pbtBuff, iLen, 1, pWP->fImage) != 1) {
      printError(ERR_WRITE_IMAGE, pWP->szImageName, NULL);
      return;
    }
    pWP->iSystemSize -= iLen;
  }

  closeCurrentFile(pWP);

  printf("\nDone. System copied.\n\n");
}
//-----------------------------------------------------------------------------
void extractSys(WORK_PARAMS *pWP)
{
  int iLen;

  if (pWP->iSystemSize == 0) {
    printError(ERR_IMG_NO_SYS, pWP->szImageName, NULL);
    return;
  }

  strcpy(pWP->szFileName, pWP->szFileMask);
  if (openCurrentFile(pWP, TRUE) != ERR_OK)
    return;

  pWP->iFileLength = pWP->iSystemSize;

  fseek(pWP->fFile, 0L, SEEK_SET);
  fseek(pWP->fImage, 0L, SEEK_SET);
  while (pWP->iFileLength > 0) {
    iLen = (pWP->iFileLength > BUFFER_SIZE) ? BUFFER_SIZE : pWP->iFileLength;
    if (fread(pWP->pbtBuff, iLen, 1, pWP->fImage) != 1) {
      printError(ERR_READ_IMAGE, pWP->szImageName, NULL);
      return;
    }

    if (fwrite(pWP->pbtBuff, iLen, 1, pWP->fFile) != 1) {
      printError(ERR_WRITE_FILE, pWP->szFileName, NULL);
      return;
    }
    pWP->iFileLength -= iLen;
  }

  closeCurrentFile(pWP);

  printf("\nDone. System extracted.\n\n");
}
//---------------------------------------------------------------------------

