//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "common.h"

//-----------------------------------------------------------------------------
void extractPathName(char *pszPath, char *pszName, char *pszExt, const char *pszSrc)
{
  char *p;

  p = strrchr(pszSrc, '/');
#if WIN
  if (p == NULL)
    p = strrchr(pszSrc, '\\');
#endif
  if (p == NULL) {
#if WIN
    *pszPath = '\0';
#else
    strcpy(pszPath, "./");
#endif
    strcpy(pszName, pszSrc);
  }
  else {
    strncpy(pszPath, pszSrc, p - pszSrc + 1);
    pszPath[p - pszSrc + 1] = '\0';
    strcpy(pszName, p + 1);
  }

  if (pszExt != NULL) {
    p = strrchr(pszName, '.');
    if (p == NULL)
      *pszExt = '\0';
    else
      strcpy(pszExt, p + 1);
  }
}
//---------------------------------------------------------------------------
void pressEnter(const char *pszMsg)
{
  if (pszMsg)
    printf(pszMsg);
  else
    printf("\nPress ENTER ...\n");
  while (getchar() != '\n')
    ;
}
//---------------------------------------------------------------------------
int printError(int iErr, const void *pvPar1, const void *pvPar2)
{
  switch (iErr) {
    case ERR_CONF_FILE_MISSING :
      fprintf(stderr, "Config file `%s.ini` missing", (char *) pvPar1);
      break;

    case ERR_CREATE_IMAGE :
      fprintf(stderr, "Image `%s` already exist", (char *) pvPar1);
      break;

    case ERR_OPEN_IMAGE :
      fprintf(stderr, "Can't open image `%s`", (char *) pvPar1);
      break;

    case ERR_IMAGE_SIZE :
      fprintf(stderr, "Image `%s` has bad size", (char *) pvPar1);
      break;

    case ERR_READ_IMAGE :
      fprintf(stderr, "Error on read from the image `%s`", (char *) pvPar1);
      break;

    case ERR_WRITE_IMAGE :
      fprintf(stderr, "Error on write to the image `%s`", (char *) pvPar1);
      break;

    case ERR_IMG_EXT_MISSING :
      fprintf(stderr, "Extension of the Image name `%s` missing", (char *) pvPar1);
      break;

    case ERR_IMG_DEF_MISSING :
      fprintf(stderr, "Definition for Image type `%s` missing", (char *) pvPar1);
      break;

    case ERR_IMG_DEF_ERROR :
      fprintf(stderr, "Error in definition for Image type `%s`", (char *) pvPar1);
      break;

    case ERR_IMG_SYS_TOO_LONG :
      fprintf(stderr, "System file `%s` is too long. Max. length is %d bytes", (char *) pvPar1, *(int *) pvPar2);
      break;

    case ERR_IMG_NO_SYS :
      fprintf(stderr, "Image `%s` does not contain system track", (char *) pvPar1);
      break;

    case ERR_IMG_BAD_TYPE :
      fprintf(stderr, "Image `%s` is of different type as main image `%s`", (char *) pvPar1, (char *) pvPar2);
      break;

    case ERR_CREATE_FILE :
      fprintf(stderr, "File `%s` already exist", (char *) pvPar1);
      break;

    case ERR_OPEN_FILE :
      fprintf(stderr, "Can't open file `%s`", (char *) pvPar1);
      break;

    case ERR_READ_FILE :
      fprintf(stderr, "Error on read from the file `%s`", (char *) pvPar1);
      break;

    case ERR_WRITE_FILE :
      fprintf(stderr, "Error on write to the file `%s`", (char *) pvPar1);
      break;

    case ERR_DISK_FULL :
      fprintf(stderr, "Disk full");
      break;

    case ERR_DIRECTORY_FULL :
      fprintf(stderr, "Directory full");
      break;

    case ERR_FILE_EXIST :
      fprintf(stderr, "File `%s` already exist", (char *) pvPar1);
      break;

    case ERR_FILE_NOT_EXIST :
      fprintf(stderr, "File `%s` not found", (char *) pvPar1);
      break;

    case ERR_BAD_FILE_NAME :
      fprintf(stderr, "Name of this file `%s` has incorrect format", (char *) pvPar1);
      break;

    case ERR_DIR_ENTRY :
      fprintf(stderr, "Corrupted directory entry `%s` (%s)", (char *) pvPar1, (char *) pvPar2);
      break;

    case ERR_FILE_TOO_LONG :
      fprintf(stderr, "File `%s` is too long (%ld bytes). Max length is 8MB.", (char *) pvPar1, *(long int *) pvPar2);
      break;

    case ERR_PATH_NOT_ALLOWED :
      fprintf(stderr, "Path not allowed for option `%s`", (char *) pvPar1);
      break;

    case ERR_FATAL_ERROR :
      fprintf(stderr, "Fatal error");
      break;

    case ERR_OUT_OF_MEMORY :
      fprintf(stderr, "Out of memory");
      break;

    case ERR_UNKNOWN_OPTION :
      fprintf(stderr, "Unknown option: `%s`", pvPar1);
      break;

    case ERR_NO_OPTION_VALUE :
      fprintf(stderr, "Missing option value(s) '%s'!", (char *) pvPar1);
      break;

    case ERR_BAD_OPTION_VALUE :
      fprintf(stderr, "Bad value `%s` for option `%s`", (char *) pvPar2, (char *) pvPar1);
      break;

    case ERR_BAD_OPTIONS :
      fprintf(stderr, "Bad combination of options: %s %s", pvPar1, pvPar2);
      break;
  }

  if (iErr != ERR_OK)
    fprintf(stderr, " !!!");
  fprintf(stderr, "\n");

  return iErr;
}
//---------------------------------------------------------------------------
int allocBuffer(void **ppvBuff, int iSize)
{
  if (*ppvBuff)
    return ERR_OK;

  *ppvBuff = (void *) malloc(iSize);
  if (!*ppvBuff)
    return printError(ERR_OUT_OF_MEMORY, NULL, NULL);

  memset(*ppvBuff, 0, iSize);
  return ERR_OK;
}
//---------------------------------------------------------------------------
int parseNumber(const char *pszStr)
{
  char *p;
  int iVal = strtol(pszStr, &p, 10);
  if (iVal < 0 || (iVal == 0 && *p != '\0'))
    iVal = -1;
  return iVal;
}
//---------------------------------------------------------------------------
char* trimLeft(char *pszStr)
{
  char *p, *q;

  if (isspace(*pszStr)) {
    p = pszStr;
    while (isspace(*p))
      p++;

    q = pszStr;
    while (*p)
      *q++ = *p++;
    *q = '\0';
  }

  return pszStr;
}
//---------------------------------------------------------------------------
char* trimRight(char *pszStr)
{
  char *p;

  p = pszStr + strlen(pszStr);
  while (p > pszStr) {
    p--;
    if (isspace(*p))
      *p = '\0';
    else
      break;
  }

  return pszStr;
}
//---------------------------------------------------------------------------
char* trim(char *pszStr)
{
  return trimLeft(trimRight(pszStr));
}
//---------------------------------------------------------------------------

